from PIL import Image
import numpy as np
import torch
from torchvision import datasets, transforms

# to use with imagenet trained NNs
IMAGENET_MEAN = (0.485, 0.456, 0.406)
IMAGENET_STD = (0.229, 0.224, 0.225)

DEFAULT_MEAN = (0.5, 0.5, 0.5)
DEFAULT_STD = (0.5, 0.5, 0.5)

 # downsize image res to 256, crop the center 224, and grayscale
PREPROCESS_TRANSFORM = transforms.Compose([
    transforms.Resize((244, 244)),
    transforms.ToTensor(),
    transforms.Normalize(DEFAULT_MEAN, DEFAULT_STD)
])

# preprocess and rotate and flip images for more data
DATA_AUGMENT_TRANSFORM = transforms.Compose([
    transforms.Resize((244, 244)),
    # careful with random rotations, they keep zeroing out the tensor...
    # transforms.RandomRotation(10),
    transforms.RandomHorizontalFlip(),
    transforms.ToTensor(),
    transforms.Normalize(DEFAULT_MEAN, DEFAULT_STD)
])

# process the image in the same way as our training data
def process_image(image_path):
    image = Image.open(image_path)
    
    transformed_image = PREPROCESS_TRANSFORM(image)
    array = np.array(transformed_image)
    
    return array



import os
import sys
import json

# get list of classes and class-to-index object from folder names
def get_classes(folder_path):
    classes = os.listdir(folder_path)
    classes.sort()
    class_to_idx = {classes[i]: i for i in range(len(classes))}
    return classes, class_to_idx

# if this file is run directly, print get_classes and output it
if (__name__ == '__main__'):
    if (len(sys.argv) != 2):
        print('Expected 1 argument: "folder_path"')
        sys.exit()

    classes, class_to_idx = get_classes(sys.argv[1])
    print(classes, class_to_idx)

    with open('classes.json', 'w') as fout:
        json.dump(classes, fout)
    
    with open('class_to_idx.json', 'w') as fout:
        json.dump(class_to_idx, fout)



import numpy as np
import torch

from torchvision import datasets, models
from ml.util import process

# force matplotlib to not use Agg
import matplotlib
matplotlib.use('Agg')
from matplotlib import pyplot as plt

# class to load the model and process images
class Predictor:
    # function to load model
    def load_model(self, checkpoint_path):
        # load the checkpoint and use cpu
        checkpoint = torch.load(checkpoint_path, map_location=torch.device('cpu'))

        # load resnet34 for transfer learning
        model = models.resnet34(pretrained=True)

        # get input features and classes
        num_input_features = model.fc.in_features
        model.class_to_idx = checkpoint['class_to_idx']
        num_classes = len(model.class_to_idx)

        # define the fully-connected layer to be a linear transformation from num_input_features to num_classes
        model.fc = torch.nn.Linear(num_input_features, num_classes)

        # load the weights from the checkpoint
        model.load_state_dict(checkpoint['state_dict'], strict=False)
    
        return model

    def __init__(self, checkpoint_path):
        # load model and use cpu
        self._model = self.load_model(checkpoint_path).cpu()
        self._classes = list(self._model.class_to_idx.keys())
    
    @property
    def classes(self):
        return self._classes

    '''
    function to predict the class of an image

    @params
    image_path = path to the image
    topk = how many of the top-predicted classes do you want returned

    @outputs
    pred_confs = np array of confidence values of predictions
    pred_classes = corresponding np aray of predicted class indexes
    '''
    def predict(self, image_path, topk=5):
        image = process.process_image(image_path)

        # convert image to float tensor and use cpu
        img_tensor = torch.from_numpy(image).type(torch.FloatTensor)
        img_tensor.cpu()

        # add a dimension to image to comply with (B x C x W x H) input of model
        img_tensor = img_tensor.unsqueeze_(0)

        # set model to evaluation mode and forward the image with gradients off
        self._model.eval()
        with torch.no_grad():
            output = self._model.forward(img_tensor)
 
        pred_confs = output.topk(topk)[0]
        pred_classes = output.topk(topk)[1]
        
        pred_confs = np.array(pred_confs)[0]
        pred_classes = np.array(pred_classes)[0]

        return pred_confs, pred_classes

    def plot_predictions(self, pred_confs, pred_classes, image_path):
        length = len(pred_classes)
        
        # convert classes to names and get prediction probabilities
        sum_confs = sum(pred_confs)
        pred_probs = []
        names = []
        for i in range(length):
            pred_probs.append(pred_confs[i] / sum_confs * 100)
            names.append(self._classes[pred_classes[i]])

        # create the plot
        plt.figure()
        plt.xlabel('Confidence Percentage')

        y_ticks = np.arange(len(names))
        plt.barh(y_ticks, pred_probs, color='blue')
        plt.yticks(y_ticks, names)
        plt.gca().invert_yaxis()
        
        plt.tight_layout()
        plt.savefig(image_path)



import sys

# allow imports from root
sys.path.append('..')

from ml import predictor

checkpoint_path = 'model/checkpoint.pth'

if (len(sys.argv) != 2):
    print('Expected 1 argument: image_path')
    sys.exit()
image_path = sys.argv[1]

predictor = predictor.Predictor(checkpoint_path)

classes = predictor.classes

num_classes = len(classes)
print(f'{num_classes} classes')

# optionally, pass in topk=?
pred_confs, pred_classes = predictor.predict(image_path, topk=10)

print('Predicted Confidences')
print(pred_confs)
print('Predicted Classes')
print(pred_classes)
print('Top Predicted Class')
print(classes[pred_classes[0]])

plot_image_path = 'plot.png'
predictor.plot_predictions(pred_confs, pred_classes, plot_image_path)

print(f'Confidence plot saved to {plot_image_path}')
